# (C) Copyright 2020-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from .constants import (
    ComparisonMode as ComparisonMode,
    DefaultValue as DefaultValue,
    TraitKind as TraitKind,
    ValidateTrait as ValidateTrait,
    NO_COMPARE as NO_COMPARE,
    OBJECT_IDENTITY_COMPARE as OBJECT_IDENTITY_COMPARE,
    RICH_COMPARE as RICH_COMPARE,
)

from .trait_errors import (
    TraitError as TraitError,
    TraitNotificationError as TraitNotificationError,
    DelegationError as DelegationError,
)

from .traits import (
    Color as Color,
    Default as Default,
    Font as Font,
    Property as Property,
    RGBColor as RGBColor,
    Trait as Trait
)

from .ctrait import CTrait as CTrait

from .has_traits import (
    ABCHasStrictTraits as ABCHasStrictTraits,
    ABCHasTraits as ABCHasTraits,
    ABCMetaHasTraits as ABCMetaHasTraits,
    AbstractViewElement as AbstractViewElement,
    HasTraits as HasTraits,
    HasStrictTraits as HasStrictTraits,
    HasPrivateTraits as HasPrivateTraits,
    HasRequiredTraits as HasRequiredTraits,
    Interface as Interface,
    SingletonHasTraits as SingletonHasTraits,
    SingletonHasStrictTraits as SingletonHasStrictTraits,
    SingletonHasPrivateTraits as SingletonHasPrivateTraits,
    MetaHasTraits as MetaHasTraits,
    Vetoable as Vetoable,
    VetoableEvent as VetoableEvent,
    observe as observe,
    on_trait_change as on_trait_change,
    cached_property as cached_property,
    property_depends_on as property_depends_on,
    provides as provides,
    isinterface as isinterface,
)

from .trait_types import (
    Any as Any,
    Int as Int,
    Float as Float,
    Complex as Complex,
    Str as Str,
    Title as Title,
    Bytes as Bytes,
    Bool as Bool,
    CInt as CInt,
    CFloat as CFloat,
    CComplex as CComplex,
    CStr as CStr,
    CBytes as CBytes,
    CBool as CBool,
    String as String,
    Regex as Regex,
    Code as Code,
    HTML as HTML,
    Password as Password,
    Callable as Callable,
    This as This,
    self as self,
    Function as Function,
    Method as Method,
    Module as Module,
    Python as Python,
    ReadOnly as ReadOnly,
    Disallow as Disallow,
    Constant as Constant,
    Delegate as Delegate,
    DelegatesTo as DelegatesTo,
    PrototypedFrom as PrototypedFrom,
    Expression as Expression,
    PythonValue as PythonValue,
    File as File,
    Directory as Directory,
    Range as Range,
    Enum as Enum,
    Tuple as Tuple,
    List as List,
    CList as CList,
    PrefixList as PrefixList,
    Set as Set,
    CSet as CSet,
    Dict as Dict,
    Map as Map,
    PrefixMap as PrefixMap,
    Instance as Instance,
    AdaptedTo as AdaptedTo,
    AdaptsTo as AdaptsTo,
    Event as Event,
    Button as Button,
    ToolbarButton as ToolbarButton,
    Either as Either,
    Union as Union,
    Type as Type,
    Subclass as Subclass,
    Symbol as Symbol,
    WeakRef as WeakRef,
    Date as Date,
    Datetime as Datetime,
    Time as Time,
    Supports as Supports,
)

# Deprecated TraitType subclasses and instances.

from .trait_types import (
    BaseUnicode as BaseUnicode,
    Unicode as Unicode,
    BaseCUnicode as BaseCUnicode,
    CUnicode as CUnicode,
    BaseLong as BaseLong,
    Long as Long,
    BaseCLong as BaseCLong,
    CLong as CLong,
    false as false,
    true as true,
    undefined as undefined,
    ListInt as ListInt,
    ListFloat as ListFloat,
    ListStr as ListStr,
    ListUnicode as ListUnicode,
    ListComplex as ListComplex,
    ListBool as ListBool,
    ListFunction as ListFunction,
    ListMethod as ListMethod,
    ListThis as ListThis,
    DictStrAny as DictStrAny,
    DictStrStr as DictStrStr,
    DictStrInt as DictStrInt,
    DictStrFloat as DictStrFloat,
    DictStrBool as DictStrBool,
    DictStrList as DictStrList,
)

from .trait_types import (
    BaseCallable as BaseCallable,
    BaseInt as BaseInt,
    BaseFloat as BaseFloat,
    BaseComplex as BaseComplex,
    BaseStr as BaseStr,
    BaseBytes as BaseBytes,
    BaseBool as BaseBool,
    BaseCInt as BaseCInt,
    BaseCFloat as BaseCFloat,
    BaseCComplex as BaseCComplex,
    BaseCStr as BaseCStr,
    BaseCBool as BaseCBool,
    BaseFile as BaseFile,
    BaseDirectory as BaseDirectory,
    BaseRange as BaseRange,
    BaseEnum as BaseEnum,
    BaseTuple as BaseTuple,
    BaseInstance as BaseInstance,
)

from .trait_types import (
    UUID as UUID,
    ValidatedTuple as ValidatedTuple
)

from .base_trait_handler import BaseTraitHandler as BaseTraitHandler
from .trait_handler import TraitHandler as TraitHandler
from .trait_type import TraitType as TraitType
from .trait_handlers import (
    TraitCoerceType as TraitCoerceType,
    TraitCastType as TraitCastType,
    TraitInstance as TraitInstance,
    TraitFunction as TraitFunction,
    TraitEnum as TraitEnum,
    TraitPrefixList as TraitPrefixList,
    TraitMap as TraitMap,
    TraitPrefixMap as TraitPrefixMap,
    TraitCompound as TraitCompound,
    TraitList as TraitList,
    TraitDict as TraitDict,
    TraitTuple as TraitTuple,
)

from .trait_numeric import (
    Array as Array,
    ArrayOrNone as ArrayOrNone,
    CArray as CArray,
)



# (C) Copyright 2020-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from typing import Any, Optional

PasswordEditors: Any
MultilineTextEditors: Any
BytesEditors: Any
SourceCodeEditor: Any
HTMLTextEditor: Any
PythonShellEditor: Any
DateEditor: Any
TimeEditor: Any
logger: Any

def password_editor(auto_set: bool = ..., enter_set: bool = ...): ...
def multi_line_text_editor(auto_set: bool = ..., enter_set: bool = ...): ...
def bytes_editor(auto_set: bool = ..., enter_set: bool = ..., encoding: Optional[Any] = ...): ...
def code_editor(): ...
def html_editor(): ...
def shell_editor(): ...
def time_editor(): ...
def date_editor(): ...
def datetime_editor(): ...
def list_editor(trait: Any, handler: Any): ...



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Test cases for weakref (WeakRef) traits. """

import contextlib
import gc
import unittest

from traits.has_traits import HasTraits
from traits.trait_types import Str, WeakRef
from traits.testing.unittest_tools import UnittestTools


class Eggs(HasTraits):
    name = Str


class Spam(HasTraits):
    eggs = WeakRef(Eggs)


@contextlib.contextmanager
def restore_gc_state():
    """Ensure that gc state is restored on exit of the with statement."""
    originally_enabled = gc.isenabled()
    try:
        yield
    finally:
        if originally_enabled:
            gc.enable()
        else:
            gc.disable()


class TestWeakRef(UnittestTools, unittest.TestCase):
    """ Test cases for weakref (WeakRef) traits. """

    def test_set_and_get(self):
        eggs = Eggs(name="platypus")
        spam = Spam()
        self.assertIsNone(spam.eggs)
        spam.eggs = eggs
        self.assertIs(spam.eggs, eggs)
        del eggs
        self.assertIsNone(spam.eggs)

    def test_target_freed_notification(self):
        eggs = Eggs(name="duck")
        spam = Spam(eggs=eggs)

        # Removal of the last reference to 'eggs' should trigger notification.
        with self.assertTraitChanges(spam, "eggs"):
            del eggs

    def test_weakref_trait_doesnt_leak_cycles(self):
        eggs = Eggs(name="ostrich")
        with restore_gc_state():
            gc.disable()
            gc.collect()
            spam = Spam(eggs=eggs)
            del spam
            self.assertEqual(gc.collect(), 0)



# (C) Copyright 2020-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from typing import (
    Any as _Any,
    Callable as _CallableType,
    Dict as _Dict,
    Type as _Type,
    Union as _Union,
)
from .trait_handler import TraitHandler


class TraitCoerceType(TraitHandler):
    def __init__(self,
                 atype: _Union[_Type, str] = ...,
                 ) -> None:
        ...


class TraitCastType(TraitCoerceType):
    ...


class TraitInstance(TraitHandler):
    def __init__(self,
                 aclass: _Union[_Type, str] = ...,
                 allow_none: bool = ...,
                 module: str = ...,
                 ) -> None:
        ...


class TraitFunction(TraitHandler):
    def __init__(self,
                 afunc: _CallableType = ...) -> None:
        ...


class TraitEnum(TraitHandler):
    def __init__(self,
                 afunc: _CallableType = ...) -> None:
        ...


class TraitPrefixList(TraitHandler):
    def __init__(self,
                 *values: _Any) -> None:
        ...


class TraitMap(TraitHandler):
    def __init__(self,
                 map: _Dict = ...) -> None:
        ...


class TraitPrefixMap(TraitMap):
    ...


class TraitCompound(TraitHandler):
    def __init__(self,
                 *handlers: _Any) -> None:
        ...


class TraitTuple(TraitHandler):
    def __init__(self,
                 *args: _Any) -> None:
        ...


class TraitList(TraitHandler):
    def __init__(self,
                 trait: _Any = ...,
                 minlen: int = ...,
                 maxlen: int = ...,
                 has_items: bool = ...,
                 ) -> None:
        ...


class TraitDict(TraitHandler):
    def __init__(self,
                 key_trait: _Any = ...,
                 value_trait: _Any = ...,
                 has_items: bool = ...,
                 ) -> None:
        ...



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# custom_traithandler.py --- Example of a custom TraitHandler

# --[Imports]------------------------------------------------------------------
from traits.api import TraitHandler


# --[Code]---------------------------------------------------------------------
class TraitOddInteger(TraitHandler):
    def validate(self, object, name, value):
        if isinstance(value, int) and (value > 0) and ((value % 2) == 1):
            return value
        self.error(object, name, value)

    def info(self):
        return "a positive odd integer"



# (C) Copyright 2020-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
A selection of code snippets used to test completeness of the stubs.
"""

from traits.api import HasStrictTraits, Str, TraitError


class HasName(HasStrictTraits):
    name = Str()


def try_assigning_age(x: HasName, new_name: str):
    try:
        x.age = new_name
    except TraitError:
        raise ValueError(f"Bad age: {new_name}")



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
This example shows how to set up notifications for changes on an object that
is an item in a list.

Note that when a new list is assigned, the change event is of a different type
compared to when a list is mutated.
"""

from traits.api import HasTraits, Int, List, observe
from traits.observation.api import trait


class Person(HasTraits):

    scores = List(Int)

    @observe("scores")
    def notify_scores_change(self, event):
        print(
            "{event.name} changed from {event.old} to {event.new}. "
            "(Event type: {event.__class__.__name__})".format(event=event)
        )

    @observe(trait("scores", notify=False).list_items())
    def notify_scores_content_change(self, event):
        print(
            "scores added: {event.added}. scores removed: {event.removed} "
            "(Event type: {event.__class__.__name__})".format(event=event)
        )


person = Person(scores=[1, 2])
# print: scores changed from [] to [1, 2]. (Event type: TraitChangeEvent)
person.scores.append(3)
# print: scores added: [3]. scores removed: [] (Event type: ListChangeEvent)



# (C) Copyright 2020-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from . import ctraits as ctraits
from .constants import ComparisonMode as ComparisonMode, DefaultValue as DefaultValue, default_value_map as default_value_map
from .trait_base import SequenceTypes as SequenceTypes, Undefined as Undefined
from .trait_dict_object import TraitDictObject as TraitDictObject
from .trait_list_object import TraitListObject as TraitListObject
from .trait_set_object import TraitSetObject as TraitSetObject
from typing import Any

def __newobj__(cls, *args: Any): ...

class CTrait(ctraits.cTrait):
    def __call__(self, *args: Any, **metadata: Any): ...
    @property
    def default(self): ...
    @property
    def default_kind(self): ...
    @property
    def trait_type(self): ...
    @property
    def inner_traits(self): ...
    @property
    def comparison_mode(self): ...
    @comparison_mode.setter
    def comparison_mode(self, value: Any) -> None: ...
    def is_trait_type(self, trait_type: Any): ...
    editor: Any = ...
    def get_editor(self): ...
    def get_help(self, full: bool = ...): ...
    def full_info(self, object: Any, name: Any, value: Any): ...
    def info(self): ...
    def as_ctrait(self): ...
    def __reduce_ex__(self, protocol: Any): ...



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


import unittest

from traits.api import Constant, HasTraits, TraitError


class TestConstantTrait(unittest.TestCase):

    def test_initial_value(self):
        class TestClass(HasTraits):
            c_atr = Constant(5)

        self.assertEqual(TestClass().c_atr, 5)

    def test_mutable_initial_value(self):
        class TestClass(HasTraits):
            c_atr_1 = Constant([1, 2, 3, 4, 5])
            c_atr_2 = Constant({"a": 1, "b": 2})

        obj = TestClass()

        self.assertEqual(obj.c_atr_1, [1, 2, 3, 4, 5])
        self.assertEqual(obj.c_atr_2, {"a": 1, "b": 2})

    def test_assign_fails(self):
        class TestClass(HasTraits):
            c_atr = Constant(5)

        with self.assertRaises(TraitError):
            TestClass(c_atr=5)
        with self.assertRaises(TraitError):
            del TestClass().c_atr

    def test_mutate_succeeds(self):
        class TestClass(HasTraits):
            c_atr_1 = Constant([1, 2, 3, 4, 5])
            c_atr_2 = Constant({"a": 1, "b": 2})

        obj = TestClass()
        obj.c_atr_1.append(6)
        obj.c_atr_2["c"] = 3

        self.assertEqual(obj.c_atr_1, [1, 2, 3, 4, 5, 6])
        self.assertEqual(obj.c_atr_2, {"a": 1, "b": 2, "c": 3})

    def test_mutate_affects_all_instances(self):
        # Mutable values are allowed for the 'Constant' trait because it's
        # impractical to do otherwise - many things are mutable but not
        # intended to be mutated, and there's no practical way to test for
        # "not-intended-to-be-mutated". Nevertheless, actually mutating the
        # value for 'Constant' is inadvisable in practice, despite the
        # existence of this test.
        class TestClass(HasTraits):
            c_atr = Constant([1, 2, 3, 4, 5])

        obj1 = TestClass()
        obj2 = TestClass()

        # Mutate obj2; check that obj1 is affected.
        obj2.c_atr.append(6)
        self.assertEqual(obj1.c_atr, [1, 2, 3, 4, 5, 6])

        # Check directly that both refer to the same object.
        self.assertIs(obj1.c_atr, obj2.c_atr)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Base classes for adapters.

Adapters do not have to inherit from these classes, as long as their
constructor takes the object to be adapted as the first and only
*positional* argument.

"""


from traits.has_traits import HasTraits
from traits.trait_types import Any


class PurePythonAdapter(object):
    """ Base class for pure Python adapters. """

    def __init__(self, adaptee):
        self.adaptee = adaptee


class Adapter(HasTraits):
    """ Base class for adapters with traits. """

    def __init__(self, adaptee, **traits):
        traits["adaptee"] = adaptee
        super().__init__(**traits)

    adaptee = Any



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import datetime
from traits.api import HasTraits, Date, Int


class TestClass(HasTraits):
    t = Date()
    x = Int()


obj = TestClass()
obj.t = datetime.datetime.now().date()
obj.t = datetime.datetime.now()
obj.t = None

obj.t = datetime.datetime.now().time()  # E: assignment
obj.t = "sometime-string"  # E: assignment
obj.t = 9  # E: assignment
obj.t = []  # E: assignment



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Tests for the Datetime trait type.
"""

import datetime
import unittest

from traits.testing.optional_dependencies import requires_traitsui, traitsui

from traits.api import Datetime, HasStrictTraits, TraitError


#: Unix epoch
UNIX_EPOCH = datetime.datetime(1970, 1, 1, 0, 0)

#: Windows NT epoch
NT_EPOCH = datetime.datetime(1600, 1, 1, 0, 0)


class HasDatetimeTraits(HasStrictTraits):
    #: Simple case - no default, no parameters, no metadata
    simple_datetime = Datetime()

    #: Datetime with default
    epoch = Datetime(UNIX_EPOCH)

    #: Datetime with default provided via keyword.
    alternative_epoch = Datetime(default_value=NT_EPOCH)

    #: None prohibited
    none_prohibited = Datetime(allow_none=False)

    #: None allowed
    none_allowed = Datetime(allow_none=True)


class TestDatetime(unittest.TestCase):
    def test_default(self):
        obj = HasDatetimeTraits()
        self.assertEqual(obj.simple_datetime, None)
        self.assertEqual(obj.epoch, UNIX_EPOCH)
        self.assertEqual(obj.alternative_epoch, NT_EPOCH)

    def test_assign_datetime(self):
        # By default, datetime instances are permitted.
        test_datetime = datetime.datetime(1975, 2, 13)
        obj = HasDatetimeTraits()
        obj.simple_datetime = test_datetime
        self.assertEqual(obj.simple_datetime, test_datetime)

    def test_assign_non_datetime(self):
        obj = HasDatetimeTraits()
        with self.assertRaises(TraitError) as exception_context:
            obj.simple_datetime = "2021-02-05 12:00:00"
        message = str(exception_context.exception)
        self.assertIn("must be a datetime or None, but", message)

    def test_assign_date(self):
        obj = HasDatetimeTraits()
        with self.assertRaises(TraitError) as exception_context:
            obj.simple_datetime = datetime.date(1975, 2, 13)
        message = str(exception_context.exception)
        self.assertIn("must be a datetime or None, but", message)
        self.assertIsNone(obj.simple_datetime)

    def test_assign_none_with_allow_none_not_given(self):
        obj = HasDatetimeTraits(simple_datetime=UNIX_EPOCH)
        self.assertIsNotNone(obj.simple_datetime)
        with self.assertWarns(DeprecationWarning) as warnings_cm:
            obj.simple_datetime = None
        self.assertIsNone(obj.simple_datetime)

        _, _, this_module = __name__.rpartition(".")
        self.assertIn(this_module, warnings_cm.filename)
        self.assertIn(
            "None will no longer be accepted",
            str(warnings_cm.warning),
        )

    def test_assign_none_with_allow_none_false(self):
        obj = HasDatetimeTraits(none_prohibited=UNIX_EPOCH)
        with self.assertRaises(TraitError) as exception_context:
            obj.none_prohibited = None
        message = str(exception_context.exception)
        self.assertIn("must be a datetime, but", message)

    def test_assign_none_with_allow_none_true(self):
        obj = HasDatetimeTraits(none_allowed=UNIX_EPOCH)
        self.assertIsNotNone(obj.none_allowed)
        obj.none_allowed = None
        self.assertIsNone(obj.none_allowed)

    @requires_traitsui
    def test_get_editor(self):
        obj = HasDatetimeTraits()
        trait = obj.base_trait("epoch")
        editor_factory = trait.get_editor()
        self.assertIsInstance(editor_factory, traitsui.api.DatetimeEditor)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# instance_trait_defaults.py --- Example of Instance trait default values

# --[Imports]------------------------------------------------------------------
from traits.api import HasTraits, Instance


# --[Code]---------------------------------------------------------------------
class Parent(HasTraits):
    pass


class Child(HasTraits):
    # default value is None
    father = Instance(Parent)
    # default value is still None, but None can not be assigned
    grandfather = Instance(Parent, allow_none=False)
    # default value is Parent()
    mother = Instance(Parent, args=())



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from contextlib import contextmanager
import unittest

from traits.ctrait import CTrait
from traits.trait_converters import (
    as_ctrait,
    check_trait,
    trait_cast,
    trait_for,
    trait_from,
)
from traits.trait_factory import TraitFactory
from traits.trait_handlers import TraitCastType, TraitInstance
from traits.api import Any, Int


@contextmanager
def reset_trait_factory():
    from traits import trait_factory
    old_tfi = trait_factory._trait_factory_instances.copy()
    try:
        yield
    finally:
        trait_factory._trait_factory_instances = old_tfi


class TestTraitCast(unittest.TestCase):

    def test_trait_cast_ctrait(self):
        ct = Int().as_ctrait()

        result = trait_cast(ct)

        self.assertIs(result, ct)

    def test_trait_cast_trait_type_class(self):
        result = trait_cast(Int)

        self.assertIsInstance(result, CTrait)
        self.assertIsInstance(result.handler, Int)

    def test_trait_cast_trait_type_instance(self):
        trait = Int()

        result = trait_cast(trait)

        self.assertIsInstance(result, CTrait)
        self.assertIs(result.handler, trait)

    def test_trait_cast_trait_factory(self):
        int_trait_factory = TraitFactory(lambda: Int().as_ctrait())

        with reset_trait_factory():
            result = trait_cast(int_trait_factory)
            ct = int_trait_factory.as_ctrait()

        self.assertIsInstance(result, CTrait)
        self.assertIs(result, ct)

    def test_trait_cast_none(self):
        result = trait_cast(None)

        self.assertIsNone(result)

    def test_trait_cast_other(self):
        result = trait_cast(1)

        self.assertIsNone(result)


class TestTraitFrom(unittest.TestCase):

    def test_trait_from_ctrait(self):
        ct = Int().as_ctrait()

        result = trait_from(ct)

        self.assertIs(result, ct)

    def test_trait_from_trait_type_class(self):
        result = trait_from(Int)

        self.assertIsInstance(result, CTrait)
        self.assertIsInstance(result.handler, Int)

    def test_trait_from_trait_type_instance(self):
        trait = Int()

        result = trait_from(trait)

        self.assertIsInstance(result, CTrait)
        self.assertIs(result.handler, trait)

    def test_trait_from_trait_factory(self):
        int_trait_factory = TraitFactory(lambda: Int().as_ctrait())

        with reset_trait_factory():
            result = trait_from(int_trait_factory)
            ct = int_trait_factory.as_ctrait()

        self.assertIsInstance(result, CTrait)
        self.assertIs(result, ct)

    def test_trait_from_none(self):
        result = trait_from(None)

        self.assertIsInstance(result, CTrait)
        self.assertIsInstance(result.handler, Any)

    def test_trait_from_other(self):
        result = trait_from(1)

        self.assertIsInstance(result, CTrait)
        self.assertIsInstance(result.handler, TraitCastType)
        self.assertEqual(result.handler.aType, int)


class TestCheckTrait(unittest.TestCase):

    def test_check_trait_ctrait(self):
        ct = Int().as_ctrait()

        result = check_trait(ct)

        self.assertIs(result, ct)

    def test_check_trait_trait_type_class(self):
        result = check_trait(Int)

        self.assertIsInstance(result, CTrait)
        self.assertIsInstance(result.handler, Int)

    def test_check_trait_trait_type_instance(self):
        trait = Int()

        result = check_trait(trait)

        self.assertIsInstance(result, CTrait)
        self.assertIs(result.handler, trait)

    def test_check_trait_trait_factory(self):
        int_trait_factory = TraitFactory(lambda: Int().as_ctrait())

        with reset_trait_factory():
            result = check_trait(int_trait_factory)
            ct = int_trait_factory.as_ctrait()

        self.assertIsInstance(result, CTrait)
        self.assertIs(result, ct)

    def test_check_trait_none(self):
        result = check_trait(None)

        self.assertIsNone(result)

    def test_check_trait_other(self):
        result = check_trait(1)

        self.assertEqual(result, 1)


class TestTraitFor(unittest.TestCase):

    def test_trait_for_ctrait(self):
        ct = Int().as_ctrait()

        result = trait_for(ct)

        self.assertIs(result, ct)

    def test_trait_for_trait_type_class(self):
        result = trait_for(Int)

        self.assertIsInstance(result, CTrait)
        self.assertIsInstance(result.handler, Int)

    def test_trait_for_trait_type_instance(self):
        trait = Int()

        result = trait_for(trait)

        self.assertIsInstance(result, CTrait)
        self.assertIs(result.handler, trait)

    def test_trait_for_trait_factory(self):
        int_trait_factory = TraitFactory(lambda: Int().as_ctrait())

        with reset_trait_factory():
            result = trait_for(int_trait_factory)
            ct = int_trait_factory.as_ctrait()

        self.assertIsInstance(result, CTrait)
        self.assertIs(result, ct)

    def test_trait_for_none(self):
        result = trait_for(None)

        self.assertIsInstance(result, CTrait)
        self.assertIsInstance(result.handler, TraitInstance)
        self.assertEqual(result.handler.aClass, type(None))

    def test_trait_for_other(self):
        result = trait_for(1)

        self.assertIsInstance(result, CTrait)
        self.assertIsInstance(result.handler, TraitCastType)
        self.assertEqual(result.handler.aType, int)


class TestAsCtrait(unittest.TestCase):

    def test_as_ctrait_from_ctrait(self):
        ct = Int().as_ctrait()

        result = as_ctrait(ct)

        self.assertIs(result, ct)

    def test_as_ctrait_from_class(self):
        result = as_ctrait(Int)

        self.assertIsInstance(result, CTrait)
        self.assertIsInstance(result.handler, Int)

    def test_as_ctrait_from_instance(self):
        trait = Int()

        result = as_ctrait(trait)

        self.assertIsInstance(result, CTrait)
        self.assertIs(result.handler, trait)

    def test_as_ctrait_from_trait_factory(self):
        int_trait_factory = TraitFactory(lambda: Int().as_ctrait())

        with reset_trait_factory():
            result = as_ctrait(int_trait_factory)
            ct = int_trait_factory.as_ctrait()

        self.assertIsInstance(result, CTrait)
        self.assertIs(result, ct)

    def test_as_ctrait_raise_exception(self):
        with self.assertRaises(TypeError):
            as_ctrait(1)

        with self.assertRaises(TypeError):
            as_ctrait(int)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Integration tests between HasTraits and observe.
See tests in ``traits.observations`` for more targeted tests.
"""

import unittest

from traits.api import (
    Any,
    Bool,
    DelegatesTo,
    Dict,
    Event,
    HasTraits,
    Instance,
    Int,
    List,
    observe,
    Property,
    Set,
    Str,
    Undefined,
)
from traits.observation.api import (
    anytrait,
    pop_exception_handler,
    push_exception_handler,
    trait,
)


class TestObserveDecorator(unittest.TestCase):
    """ General tests for the observe decorator. """

    def test_warning_on_handler_with_bad_signature(self):
        message_regex = "should be callable with a single positional argument"

        with self.assertWarnsRegex(UserWarning, message_regex):
            class A(HasTraits):
                foo = Int()

                @observe("foo")
                def _do_something_when_foo_changes(self):
                    pass

        with self.assertWarnsRegex(UserWarning, message_regex):
            class B(HasTraits):
                foo = Int()

                @observe("foo")
                def _do_something_when_foo_changes(self, **kwargs):
                    pass

    def test_decorated_method_signatures(self):
        # Test different handler signatures for compatibility with
        # observe decorator.

        class A(HasTraits):
            foo = Int()

            call_count = Int(0)

            @observe("foo")
            def _the_usual_signature(self, event):
                self.call_count += 1

            @observe("foo")
            def _method_with_extra_optional_args(self, event, frombicate=True):
                self.call_count += 1

            @observe("foo")
            def _method_with_star_args(self, *args):
                self.call_count += 1

            @observe("foo")
            def _method_with_alternative_name(self, foo_change_event):
                self.call_count += 1

            @observe("foo")
            def _optional_second_argument(self, event=None):
                self.call_count += 1

        a = A()
        self.assertEqual(a.call_count, 0)
        a.foo = 23
        self.assertEqual(a.call_count, 5)


class Student(HasTraits):
    """ Model for testing list + post_init (enthought/traits#275) """

    graduate = Event()


class Teacher(HasTraits):
    """ Model for testing list + post_init (enthought/traits#275) """

    students = List(Instance(Student))

    student_graduate_events = List()

    @observe(
        trait("students", notify=True)
        .list_items(notify=False)
        .trait("graduate"),
        post_init=True)
    def _student_graduate(self, event):
        self.student_graduate_events.append(event)


class TestHasTraitsObservePostInit(unittest.TestCase):
    """ Test for enthought/traits#275 """

    def setUp(self):
        push_exception_handler(reraise_exceptions=True)
        self.addCleanup(pop_exception_handler)

    def test_observe_post_init_true(self):
        # Resolves enthought/traits#275
        students = [Student() for _ in range(3)]
        teacher = Teacher(students=students)

        # No events as handler is created post-init
        self.assertEqual(len(teacher.student_graduate_events), 0)

        # when
        students[0].graduate = True

        # then
        self.assertEqual(len(teacher.student_graduate_events), 1)


# Integration tests for default initializer -----------------------------------


class Record(HasTraits):
    number = Int(10)

    default_call_count = Int()

    number_change_events = List()

    clicked = Event()

    def _number_default(self):
        self.default_call_count += 1
        return 99

    @observe('number')
    def handle_number_change(self, event):
        self.number_change_events.append(event)


class Album(HasTraits):

    records = List(Instance(Record))

    records_default_call_count = Int()

    record_number_change_events = List()

    name_to_records = Dict(Str, Record)

    name_to_records_default_call_count = Int()

    name_to_records_clicked_events = List()

    def _records_default(self):
        self.records_default_call_count += 1
        return [Record()]

    @observe(trait("records").list_items().trait("number"))
    def handle_record_number_changed(self, event):
        self.record_number_change_events.append(event)

    def _name_to_records_default(self):
        self.name_to_records_default_call_count += 1
        return {"Record": Record()}

    @observe("name_to_records:items:clicked")
    def handle_event(self, event):
        self.name_to_records_clicked_events.append(event)


class TestHasTraitsObserverDefaultHandler(unittest.TestCase):
    """ Test the behaviour with dynamic default handler + container. """

    def setUp(self):
        push_exception_handler(reraise_exceptions=True)
        self.addCleanup(pop_exception_handler)

    def test_default_not_called_if_init_contains_value(self):
        record = Record(number=123)
        # enthought/traits#94
        self.assertEqual(record.default_call_count, 1)
        self.assertEqual(len(record.number_change_events), 1)
        event, = record.number_change_events
        self.assertEqual(event.object, record)
        self.assertEqual(event.name, "number")
        self.assertEqual(event.old, 99)
        self.assertEqual(event.new, 123)

    def test_observe_extended_trait_in_list(self):
        album = Album()

        # default is not called.
        self.assertEqual(album.records_default_call_count, 0)
        self.assertEqual(len(album.record_number_change_events), 0)

        # But the observers are hooked up
        # when
        album.records[0].number += 1

        # then
        self.assertEqual(album.records_default_call_count, 1)
        self.assertEqual(len(album.record_number_change_events), 1)
        event, = album.record_number_change_events
        self.assertEqual(event.object, album.records[0])
        self.assertEqual(event.name, "number")
        self.assertEqual(event.old, 99)
        self.assertEqual(event.new, 100)

    def test_observe_extended_trait_in_default_dict(self):
        # Test for enthought/traits#279
        album = Album()

        self.assertEqual(album.name_to_records_default_call_count, 0)
        self.assertEqual(len(album.name_to_records_clicked_events), 0)

        # when
        album.name_to_records["Record"].clicked = True

        # then
        self.assertEqual(len(album.name_to_records_clicked_events), 1)


# Integration tests for nested List and extended traits -----------------------

class SingleValue(HasTraits):

    value = Int()


class ClassWithListOfInstance(HasTraits):

    list_of_instances = List(Instance(SingleValue))


class ClassWithListOfListOfInstance(HasTraits):

    list_of_list_of_instances = List(List(Instance(SingleValue)))


class TestHasTraitsObserveListOfInstance(unittest.TestCase):

    def setUp(self):
        push_exception_handler(reraise_exceptions=True)
        self.addCleanup(pop_exception_handler)

    def test_observe_instance_in_nested_list(self):

        container = ClassWithListOfListOfInstance()
        events = []
        handler = events.append
        container.observe(
            expression=(
                trait("list_of_list_of_instances", notify=False)
                .list_items(notify=False)
                .list_items(notify=False)
                .trait("value")
            ),
            handler=handler,
        )

        # sanity check
        single_value_instance = SingleValue()
        inner_list = [single_value_instance]
        container.list_of_list_of_instances.append(inner_list)
        self.assertEqual(len(events), 0)

        # when
        single_value_instance.value += 1

        # then
        event, = events
        self.assertEqual(event.object, single_value_instance)
        self.assertEqual(event.name, "value")
        self.assertEqual(event.old, 0)
        self.assertEqual(event.new, 1)

    def test_nested_list_reassigned_value_compared_equally(self):
        container = ClassWithListOfListOfInstance()
        events = []
        handler = events.append
        container.observe(
            expression=(
                trait("list_of_list_of_instances", notify=False)
                .list_items(notify=False)
                .list_items(notify=False)
                .trait("value")
            ),
            handler=handler,
        )

        inner_list = [SingleValue()]
        container.list_of_list_of_instances = [inner_list]
        # sanity check
        self.assertEqual(len(events), 0)

        # assignment of a list that compares equally should be handled
        # correctly.
        # This relies on TraitList not trying to suppress notifications
        # when new values compared equally to old values.
        container.list_of_list_of_instances[0] = inner_list
        second_instance = SingleValue()
        container.list_of_list_of_instances[0].append(second_instance)
        self.assertEqual(len(events), 0)

        # when
        second_instance.value += 1

        # then
        event, = events
        self.assertEqual(event.object, second_instance)
        self.assertEqual(event.name, "value")
        self.assertEqual(event.old, 0)
        self.assertEqual(event.new, 1)

    def test_duplicated_items_tracked(self):
        # test for enthought/traits#237
        container = ClassWithListOfInstance()
        events = []
        handler = events.append
        container.observe(
            expression=(
                trait("list_of_instances", notify=False)
                .list_items(notify=False)
                .trait("value")
            ),
            handler=handler,
        )

        instance = SingleValue()
        # The item is repeated.
        container.list_of_instances.append(instance)
        container.list_of_instances.append(instance)
        self.assertEqual(len(events), 0)

        # when
        instance.value += 1

        # then
        self.assertEqual(len(events), 1)
        events.clear()

        # when
        container.list_of_instances.pop()
        instance.value += 1

        # then
        self.assertEqual(len(events), 1)
        events.clear()

        # when
        container.list_of_instances.pop()
        instance.value += 1

        # then
        self.assertEqual(len(events), 0)


# Integration tests for nested Dict and extended traits -----------------------


class ClassWithDictOfInstance(HasTraits):

    name_to_instance = Dict(Str, Instance(SingleValue))


class TestHasTraitsObserveDictOfInstance(unittest.TestCase):

    def setUp(self):
        push_exception_handler(reraise_exceptions=True)
        self.addCleanup(pop_exception_handler)

    def test_observe_instance_in_dict(self):
        container = ClassWithDictOfInstance()
        events = []
        handler = events.append
        container.observe(
            handler=handler,
            expression=(
                trait("name_to_instance", notify=False)
                .dict_items(notify=False)
                .trait("value")
            ),
        )

        single_value_instance = SingleValue()
        container.name_to_instance = {"name": single_value_instance}
        # sanity check
        self.assertEqual(len(events), 0)

        # when
        single_value_instance.value += 1

        # then
        event, = events
        self.assertEqual(event.object, single_value_instance)
        self.assertEqual(event.name, "value")
        self.assertEqual(event.old, 0)
        self.assertEqual(event.new, 1)


# Integration tests for Set and extended traits ------------------------------


class ClassWithSetOfInstance(HasTraits):

    instances = Set(Instance(SingleValue))

    instances_compat = Set(Instance(SingleValue))


class TestHasTraitsObserveSetOfInstance(unittest.TestCase):

    def setUp(self):
        push_exception_handler(reraise_exceptions=True)
        self.addCleanup(pop_exception_handler)

    def test_observe_instance_in_set(self):
        container = ClassWithSetOfInstance()
        events = []
        handler = events.append
        container.observe(
            handler=handler,
            expression=(
                trait("instances", notify=False)
                .set_items(notify=False)
                .trait("value")
            ),
        )

        single_value_instance = SingleValue()
        container.instances = set([single_value_instance])
        # sanity check
        self.assertEqual(len(events), 0)

        # when
        single_value_instance.value += 1

        # then
        event, = events
        self.assertEqual(event.object, single_value_instance)
        self.assertEqual(event.name, "value")
        self.assertEqual(event.old, 0)
        self.assertEqual(event.new, 1)


# Integration test for maintaining and differentiating observers --------------

class Potato(HasTraits):

    name = Str()


class PotatoBag(HasTraits):

    potatos = List(Instance(Potato))


class Crate(HasTraits):

    potato_bags = List(PotatoBag)


class TestHasTraitsObserverDifferentiateParent(unittest.TestCase):

    def test_shared_instance_but_different_target(self):
        # If the comparison of targets is removed from
        # TraitEventNotifier.equals, this test would fail.
        potato = Potato()
        potato_bag = PotatoBag(potatos=[potato])
        crate1 = Crate(potato_bags=[potato_bag])
        crate2 = Crate(potato_bags=[potato_bag])

        # when
        events = []
        handler = events.append
        crate1.observe(
            handler, "potato_bags:items:potatos:items:name",
        )
        crate2.observe(
            handler, "potato_bags:items:potatos:items:name",
        )
        potato.name = "King Edward"

        # then
        # there are two notifiers, because they are observed from different
        # objects.
        self.assertEqual(len(events), 2)

    def test_shared_instance_same_graph_different_target(self):

        crate1 = Crate()
        crate2 = Crate()

        # given
        events = []
        handler = events.append
        crate1.observe(handler, "potato_bags:items:potatos:items:name")
        crate2.observe(handler, "potato_bags:items:potatos:items:name")

        new_potato = Potato()
        new_potato_bag = PotatoBag(potatos=[new_potato])
        crate1.potato_bags = [new_potato_bag]
        crate2.potato_bags = [new_potato_bag]
        new_potato.name = "King Edward I"
        self.assertEqual(len(events), 2)
        events.clear()

        # when
        # remove the second observer
        crate2.observe(
            handler, "potato_bags:items:potatos:items:name", remove=True)
        new_potato.name = "King Edward II"

        # then
        self.assertEqual(len(events), 1)
        events.clear()

        # then
        # This check the observer is maintained.
        maris_piper = Potato()
        crate2.potato_bags[0].potatos.append(maris_piper)
        crate1.potato_bags = []
        self.assertEqual(len(events), 0)  # sanity check

        # this fails if targets were not compared.
        maris_piper.name = "Maris Piper"
        self.assertEqual(len(events), 0)


# Integration test for the special event metadata ----------------------------

class FooWithEventMetadata(HasTraits):
    val = Str(event="the_trait")

    @observe("the_trait")
    def _handle_the_trait_changed(self, event):
        pass


class TestSpecialEvent(unittest.TestCase):
    """ Test the 'event' metadata... won't work with ``observe``!
    """

    def setUp(self):
        push_exception_handler(reraise_exceptions=True)
        self.addCleanup(pop_exception_handler)

    def test_events(self):

        with self.assertRaises(ValueError) as exception_cm:
            # Attempt to attach the observer will fail because
            # the "the_trait" is not actually a trait on the object.
            FooWithEventMetadata()

        self.assertIn(
            "Trait named 'the_trait' not found",
            str(exception_cm.exception),
        )


# Integration test for when the observer is not appropriate for the data ------

class Person(HasTraits):
    name = Str()


class Team(HasTraits):
    leader = Instance(Person)

    member_names = List(Str())

    any_value = Any()


class TestObserverError(unittest.TestCase):

    def setUp(self):
        push_exception_handler(reraise_exceptions=True)
        self.addCleanup(pop_exception_handler)

    def test_trait_is_not_list(self):

        team = Team()
        # The `list_items` should not be used here.
        # Error is not emitted now as leader is not defined so there is no
        # way to check.
        team.observe(lambda e: None, trait("leader").list_items())

        person = Person()
        with self.assertRaises(ValueError) as exception_cm:
            team.leader = person

        self.assertIn(
            "Expected a TraitList to be observed",
            str(exception_cm.exception),
        )

    def test_items_on_a_list_not_observable_by_named_trait(self):
        # The member_names is a list of str, attempt to observe extended
        # trait on them should fail.
        team = Team()

        team.observe(
            lambda e: None,
            trait("member_names").list_items().trait("does_not_exist")
        )

        with self.assertRaises(ValueError) as exception_cm:
            team.member_names = ["Paul"]

        self.assertEqual(
            str(exception_cm.exception),
            "Trait named 'does_not_exist' not found on 'Paul'."
        )

    def test_extended_trait_on_any_value(self):
        team = Team()
        team.any_value = 123

        with self.assertRaises(ValueError) as exception_cm:
            team.observe(
                lambda e: None, trait("any_value").trait("does_not_exist"))

        self.assertEqual(
            str(exception_cm.exception),
            "Trait named 'does_not_exist' not found on 123."
        )

    def test_no_new_trait_added(self):
        # Test enthought/traits#447 can be avoided with observe
        team = Team()
        team.observe(lambda e: None, trait("leader").trait("does_not_exist"))

        with self.assertRaises(ValueError):
            team.leader = Person()

        self.assertNotIn("does_not_exist", team.leader.trait_names())


# Integration test with DelegateTo --------------------------------------------

class Dummy(HasTraits):
    x = Int(10)


class Dummy2(HasTraits):
    y = Int(20)
    dummy = Instance(Dummy)


class DelegateMess(HasTraits):
    dummy1 = Instance(Dummy, args=())
    dummy2 = Instance(Dummy2)

    y = DelegatesTo("dummy2")

    handler_called = Bool(False)

    def _dummy2_default(self):
        # Create `self.dummy1`
        return Dummy2(dummy=self.dummy1)

    @observe("dummy1.x")
    def _on_dummy1_x(self, event):
        self.handler_called = True


class TestDelegateToInteraction(unittest.TestCase):

    def test_delegate_initializer(self):
        mess = DelegateMess()
        self.assertFalse(mess.handler_called)
        mess.dummy1.x = 20
        self.assertTrue(mess.handler_called)


# Integration tests with on_trait_change and observe ------------------
# The legacy of on_trait_change means instance trait named with "_items"
# suffix is handled differently in HasTraits. This tests the awkward
# interaction that could arise while using on_trait_change together with
# observe involving "*_items"

class Application(HasTraits):
    pass


class TestObserveItemsFromOnTraitChange(unittest.TestCase):

    def test_observe_event_with_undefined_name_suffix_items(self):
        # Regression test for the error resulting from trying (and failing) to
        # retrieve the CTrait for an instance trait with name "*_items"
        # via HasTraits.traits
        app = Application()

        def dummy_handler():
            pass

        # on_trait_change does not check if the trait has been defined.
        # This has the side-effect of creating the CTrait for this trait name.
        app.on_trait_change(dummy_handler, "i_am_undefined_with_items")
        self.assertIsNotNone(app._trait("i_am_undefined_with_items", 0))

        # Precondition for this test, i_am_undefined_with_items is still not
        # reported by HasTraits.traits method
        self.assertNotIn("i_am_undefined_with_items", app.traits())

        events = []
        # This works because the CTrait is created by on_trait_change
        app.observe(events.append, "i_am_undefined_with_items")

        # This should not fail.
        app.trait_property_changed("i_am_undefined_with_items", 1, 2)

        self.assertEqual(len(events), 1)


# Integration tests for anytrait listener -------------------------------------

class HasVariousTraits(HasTraits):
    #: Function to call on any trait change.
    trait_change_callback = Any()

    foo = Int(16)

    bar = Str("off")

    updated = Event(Bool)

    @observe("*")
    def _record_trait_change(self, event):
        callback = self.trait_change_callback
        if callback is not None:
            callback(event)


class UpdateListener(HasTraits):
    foo = Instance(HasVariousTraits)

    bar = Instance(HasVariousTraits)


class TestObserveAnytrait(unittest.TestCase):
    def test_observe_method_anytrait(self):
        obj = HasVariousTraits()

        events = []
        obj.observe(events.append, "*")
        obj.foo = 23
        obj.bar = "on"

        self.assertEqual(len(events), 2)
        foo_event, bar_event = events

        self.assertEqual(foo_event.object, obj)
        self.assertEqual(foo_event.name, "foo")
        self.assertEqual(foo_event.old, 16)
        self.assertEqual(foo_event.new, 23)

        self.assertEqual(bar_event.object, obj)
        self.assertEqual(bar_event.name, "bar")
        self.assertEqual(bar_event.old, "off")
        self.assertEqual(bar_event.new, "on")

    def test_observe_decorator_anytrait(self):
        events = []
        obj = HasVariousTraits(trait_change_callback=events.append)
        obj.foo = 23
        obj.bar = "on"

        self.assertEqual(len(events), 3)
        callback_event, foo_event, bar_event = events

        self.assertEqual(callback_event.object, obj)
        self.assertEqual(callback_event.name, "trait_change_callback")
        self.assertIs(callback_event.old, None)
        self.assertEqual(callback_event.new, events.append)

        self.assertEqual(foo_event.object, obj)
        self.assertEqual(foo_event.name, "foo")
        self.assertEqual(foo_event.old, 16)
        self.assertEqual(foo_event.new, 23)

        self.assertEqual(bar_event.object, obj)
        self.assertEqual(bar_event.name, "bar")
        self.assertEqual(bar_event.old, "off")
        self.assertEqual(bar_event.new, "on")

    def test_anytrait_expression(self):
        obj = HasVariousTraits()

        events = []
        obj.observe(events.append, anytrait())
        obj.foo = 23
        obj.bar = "on"

        self.assertEqual(len(events), 2)
        foo_event, bar_event = events

        self.assertEqual(foo_event.object, obj)
        self.assertEqual(foo_event.name, "foo")
        self.assertEqual(foo_event.old, 16)
        self.assertEqual(foo_event.new, 23)

        self.assertEqual(bar_event.object, obj)
        self.assertEqual(bar_event.name, "bar")
        self.assertEqual(bar_event.old, "off")
        self.assertEqual(bar_event.new, "on")

    def test_anytrait_method(self):
        foo = HasVariousTraits()
        bar = HasVariousTraits()
        obj = UpdateListener(foo=foo, bar=bar)

        events = []
        obj.observe(events.append, trait("foo", notify=False).anytrait())

        foo.updated = True
        bar.updated = True

        self.assertEqual(len(events), 1)
        foo_event, = events

        self.assertEqual(foo_event.object, foo)
        self.assertEqual(foo_event.name, "updated")
        self.assertEqual(foo_event.old, Undefined)
        self.assertEqual(foo_event.new, True)

    def test_anytrait_with_children(self):
        foo = HasVariousTraits()
        bar = HasVariousTraits()
        obj = UpdateListener(foo=foo, bar=bar)

        events = []
        with self.assertRaises(ValueError):
            obj.observe(events.append, "*:updated")

    def test_anytrait_of_anytrait(self):
        foo = HasVariousTraits()
        bar = HasVariousTraits()
        obj = UpdateListener(foo=foo, bar=bar)

        events = []
        with self.assertRaises(ValueError):
            obj.observe(events.append, "*:*")

    def test_anytrait_unobserve(self):
        obj = HasVariousTraits()

        events = []
        obj.observe(events.append, "*")
        obj.foo = 23
        obj.bar = "on"

        self.assertEqual(len(events), 2)

        obj.observe(events.append, "*", remove=True)

        obj.foo = 232
        obj.bar = "mid"

        # No additional events.
        self.assertEqual(len(events), 2)

    def test_property_subclass_observe(self):
        # Regression test for enthought/traits#1586
        class Base(HasTraits):
            bar = Int()

            foo = Property(Int(), observe="bar")

            def _get_foo(self):
                return self.bar

        class Derived(Base):
            pass

        events = []

        obj = Derived(bar=3)
        obj.observe(events.append, "foo")

        # Changing bar should result in a single event.
        self.assertEqual(len(events), 0)
        obj.bar = 5
        self.assertEqual(len(events), 1)



# (C) Copyright 2005-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Test the 'add_trait_listener', 'remove_trait_listener' interface to the
HasTraits class.

"""

import contextlib
import io
import logging
import sys
import threading
import time
import unittest

from traits.api import HasTraits, Str, Int, Float, Any, Event
from traits.api import push_exception_handler, pop_exception_handler


@contextlib.contextmanager
def captured_stderr():
    """
    Return a context manager that directs all stderr output to a string.

    """
    new_stderr = io.StringIO()
    original_stderr = sys.stderr
    sys.stderr = new_stderr
    try:
        yield new_stderr
    finally:
        sys.stderr = original_stderr


class GenerateEvents(HasTraits):
    name = Str
    age = Int
    weight = Float


events = {}  # dict of events


class ListenEvents(HasTraits):

    #  'GenerateEvents' event interface:
    #  the events are stored in the dict 'events'

    def _name_changed(self, object, name, old, new):
        events["_name_changed"] = (name, old, new)

    def _age_changed(self, object, name, old, new):
        events["_age_changed"] = (name, old, new)

    def _weight_changed(self, object, name, old, new):
        events["_weight_changed"] = (name, old, new)

    def alt_name_changed(self, object, name, old, new):
        events["alt_name_changed"] = (name, old, new)

    def alt_weight_changed(self, object, name, old, new):
        events["alt_weight_changed"] = (name, old, new)


class GenerateFailingEvents(HasTraits):
    name = Str

    def _name_changed(self):
        raise RuntimeError


class TestListeners(unittest.TestCase):
    def test_listeners(self):
        global events

        # FIXME: comparing floats
        ge = GenerateEvents()
        le = ListenEvents()

        # Starting test: No Listeners
        ge.trait_set(name="Joe", age=22, weight=152.0)

        # Adding default listener
        ge.add_trait_listener(le)
        events = {}
        ge.trait_set(name="Mike", age=34, weight=178.0)
        self.assertEqual(
            events,
            {
                "_age_changed": ("age", 22, 34),
                "_weight_changed": ("weight", 152.0, 178.0),
                "_name_changed": ("name", "Joe", "Mike"),
            },
        )

        # Adding alternate listener
        ge.add_trait_listener(le, "alt")
        events = {}
        ge.trait_set(name="Gertrude", age=39, weight=108.0)
        self.assertEqual(
            events,
            {
                "_age_changed": ("age", 34, 39),
                "_name_changed": ("name", "Mike", "Gertrude"),
                "_weight_changed": ("weight", 178.0, 108.0),
                "alt_name_changed": ("name", "Mike", "Gertrude"),
                "alt_weight_changed": ("weight", 178.0, 108.0),
            },
        )

        # Removing default listener
        ge.remove_trait_listener(le)
        events = {}
        ge.trait_set(name="Sally", age=46, weight=118.0)
        self.assertEqual(
            events,
            {
                "alt_name_changed": ("name", "Gertrude", "Sally"),
                "alt_weight_changed": ("weight", 108.0, 118.0),
            },
        )

        # Removing alternate listener
        ge.remove_trait_listener(le, "alt")
        events = {}
        ge.trait_set(name="Ralph", age=29, weight=198.0)
        self.assertEqual(events, {})

    def test_trait_exception_handler_can_access_exception(self):
        """ Tests if trait exception handlers can access the traceback of the
        exception.
        """
        from traits import trait_notifiers

        def _handle_exception(obj, name, old, new):
            self.assertIsNotNone(sys.exc_info()[0])

        ge = GenerateFailingEvents()
        try:
            trait_notifiers.push_exception_handler(
                _handle_exception, reraise_exceptions=False, main=True
            )
            ge.trait_set(name="John Cleese")
        finally:
            trait_notifiers.pop_exception_handler()

    def test_exceptions_logged(self):
        # Check that default exception handling logs the exception.
        ge = GenerateFailingEvents()

        traits_logger = logging.getLogger("traits")

        with self.assertLogs(
                logger=traits_logger, level=logging.ERROR) as log_watcher:
            ge.name = "Terry Jones"

        self.assertEqual(len(log_watcher.records), 1)
        log_record = log_watcher.records[0]

        self.assertIn(
            "Exception occurred in traits notification handler",
            log_record.message,
        )

        _, exc_value, exc_traceback = log_record.exc_info
        self.assertIsInstance(exc_value, RuntimeError)
        self.assertIsNotNone(exc_traceback)


class A(HasTraits):
    exception = Any

    foo = Event

    def foo_changed_handler(self):
        pass


def foo_writer(a, stop_event):
    while not stop_event.is_set():
        try:
            a.foo = True
        except Exception as e:
            a.exception = e


class TestRaceCondition(unittest.TestCase):
    def setUp(self):
        push_exception_handler(
            handler=lambda *args: None, reraise_exceptions=True, main=True
        )

    def tearDown(self):
        pop_exception_handler()

    def test_listener_thread_safety(self):
        # Regression test for GitHub issue #56
        a = A()
        stop_event = threading.Event()

        t = threading.Thread(target=foo_writer, args=(a, stop_event))
        t.start()

        for _ in range(100):
            a.on_trait_change(a.foo_changed_handler, "foo")
            time.sleep(0.0001)  # encourage thread-switch
            a.on_trait_change(a.foo_changed_handler, "foo", remove=True)

        stop_event.set()
        t.join()

        self.assertTrue(a.exception is None)

    def test_listener_deleted_race(self):
        # Regression test for exception that occurred when the listener_deleted
        # method is called after the dispose method on a
        # TraitsChangeNotifyWrapper.
        class SlowListener(HasTraits):
            def handle_age_change(self):
                time.sleep(1.0)

        def worker_thread(event_source, start_event):
            # Wait until the listener is set up on the main thread, then fire
            # the event.
            start_event.wait()
            event_source.age = 11

        def main_thread(event_source, start_event):
            listener = SlowListener()
            event_source.on_trait_change(listener.handle_age_change, "age")
            start_event.set()
            # Allow time to make sure that we're in the middle of handling an
            # event.
            time.sleep(0.5)
            event_source.on_trait_change(
                listener.handle_age_change, "age", remove=True
            )

        # Previously, a ValueError would be raised on the worker thread
        # during (normal refcount-based) garbage collection.  That
        # ValueError is ignored by the Python system, so the only
        # visible effect is the output to stderr.
        with captured_stderr() as s:
            start_event = threading.Event()
            event_source = GenerateEvents(age=10)
            t = threading.Thread(
                target=worker_thread, args=(event_source, start_event)
            )
            t.start()
            main_thread(event_source, start_event)
            t.join()

        self.assertNotIn("Exception", s.getvalue())
